use tokio::net::{TcpSocket, TcpStream};
use std::io;
use domo_proto::packet::Packet;
use std::net::SocketAddr;
use log::trace;
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use domo_proto::packet::identifier::Identifier;
use domo_proto::packet::packet_data::PacketData;

pub struct Client {
    socket_addr: SocketAddr,
}

impl Client {
    pub fn new(socket_addr: SocketAddr) -> Self {
        Self {
            socket_addr
        }
    }

    pub async fn client_connection(self) -> io::Result<ClientConnection> {
        ClientConnection::new(self).await
    }
}

pub struct ClientConnection {
    stream: TcpStream,
}

impl ClientConnection {
    pub async fn new(c: Client) -> io::Result<Self> {
        let sock = TcpSocket::new_v4()?;
        let stream = sock.connect(c.socket_addr).await?;
        Ok(Self {
            stream
        })
    }

    pub async fn send(&mut self, packet: Packet) -> io::Result<Packet> {
        self.stream.write(packet.build_full_packet().as_slice()).await?;
        from_tcp_stream(&mut self.stream).await
    }
}

pub async fn from_tcp_stream(stream: &mut TcpStream) -> io::Result<Packet> {
    // Implement the logic to read data from the TcpStream asynchronously and create a Packet instance

    match stream.read_u8().await? {
        0x01 => {
            // Example: Read packet data and create a Packet::V1 instance
            let src = Identifier::from(stream.read_u32().await?); // Assuming you have a similar function for Identifier
            let dest = Identifier::from(stream.read_u32().await?); // Assuming you have a similar function for Identifier
            let packet_id = Identifier::from(stream.read_u32().await?); // Assuming you have a similar function for Identifier
            let reply_to = Identifier::from(stream.read_u32().await?); // Assuming you have a similar function for Identifier
            let command = stream.read_u8().await?; // Read u8 directly
            let data_length = stream.read_u16().await?; // Read u16 directly

            trace!("received => src: {src}, dest: {dest}, packet_id: {packet_id}, reply_to: {reply_to}, command: {command}, data_length: {data_length}");

            let mut data = vec![0u8; data_length as usize];
            stream.read_exact(&mut data).await?;

            // Create a Packet::V1 instance
            let packet = Packet::V1 {
                src,
                dest,
                packet_id,
                reply_to,
                command,
                data: PacketData::new(data), // Assuming PacketData has a constructor
            };

            Ok(packet)
        }
        _ => Err(io::Error::new(io::ErrorKind::InvalidData, "received invalid data."))
    }
}